# Site-O-Mat Architecture

This document describes Site-O-Mat architecture: algorithms, data structures,
decisions, and trade-offs.

## Trade-offs

- Site-O-Mat is intended for building websites of small and medium size (in
  order of 10s-100s MB) on a relatively modern hardware (multi-core CPI, at
  least 4GB RAM). Therefore, when building a website, it prefers to minimize
  file reads and directory scans by keeping all the data in RAM.

## `build` algorithm

This is the algorithm of `build` command written in pseudo-Haskell:

```
do
  (pages, structure) <- readContent
  templates <- parseTemplates

  if not (dirExists "/build")
    then mkdir "/build"
    else removeAllFrom "/build"

  mkdir "/build/static"
  copyAll {src = "/static", dst = "/build/static"}

  for_ pages $ \page ->
    let m = getMetadata page & insert "_structure" structure
        content = render {template = getContent page, params = m}
        m' = insert "_html" (commonMarkToHTML content) m
        output =
          case lookup (lookupOrDef "_template" "default" m) templates of
            Just template ->
              render {template = template, params = m'}
            Nothing ->
              content
    in writeFile {dst = "/build" </> lookupUnsafe "_path" m </> "index.html", content = output}
```

Notes:

- the last `for_` loop is a good place for parallelism

## `readContent` algorithm

```
do
  (pages, struct) <- walk "/content" Map.empty
  (pages, lookupUnsafe "_items" struct)
where
  walk path meta
    | path = "/content/static.md" || path = "/content/static/" =
      ([], Map.empty)
    | isDir path = do
      collectionMeta <-
        if fileExists (path </> "metadata.json")
          then parseJSON $ read $ path </> "metadata.json"
          else Map.empty
      children <- lsDir path
      (pages, structs) <- unzip $ map (\c -> walk c (merge meta collectionMeta)) children
      -- sort non-collection structs according to `_sort_by` and `_reverse` fields
      ( concat pages
      , Map.fromList
        [ ("_name", dirName path),
          -- insert _is_page, _is_collection params
          -- insert collectionMeta
          ("_items", concat structs)
        ]
      )
    | isFileOrLink path && hasExt ".md" path = do
      (pageMeta, pageContent) <- readPage path
      -- insert _name, _path, _is_page, _is_collection params in pageMeta
      ( [Page { metadata = pageMeta, content = pageContent}]
      , [pageMeta]
      )
    | otherwise =
      ([], Map.empty)
```

## `parseTemplates` algorithm

```
templates <-
  listDirectory "/templates"
    <&> filter (".html" `isSuffixOf` && fileExist)
map compileMustacheFile templates
```