{-# LANGUAGE ScopedTypeVariables #-}

module Build (exec) where

import Commonmark (Html, commonmark, renderHtml)
import Control.Monad (filterM)
import qualified Data.Aeson as Aeson
import qualified Data.Aeson.Key as Key
import qualified Data.Aeson.KeyMap as M
import Data.Bifunctor (first)
import qualified Data.ByteString as BS
import Data.Foldable (traverse_)
import qualified Data.List as L
import qualified Data.Map.Strict as Map
import Data.Ord (Down (Down))
import qualified Data.Text.Lazy as LT
import Data.Time (getCurrentTime)
import Data.Time.Format.ISO8601 (iso8601Show)
import qualified Data.Vector as V
import OkStd
import System.Directory (copyFile, createDirectory, createDirectoryIfMissing, doesDirectoryExist, doesFileExist, listDirectory, removePathForcibly)
import System.FilePath (dropExtension, dropFileName, makeRelative, takeBaseName, takeFileName, (</>))
import Text.Microstache (PName (..), Template (..), compileMustacheFile, compileMustacheText, renderMustache)

exec path =
  doesDirectoryExist path >>= \case
    False ->
      error $ "directory does not exist: " <> toText path
    True -> do
      (struct, content) <- readContent path
      templates <- parseTemplates path
      removePathForcibly buildDir >> createDirectory buildDir
      copyRecursively (path </> "static") (buildDir </> "static")
      for_ content $ \case
        (Page filePath pageMeta pageContent) ->
          buildPage struct templates filePath pageMeta pageContent
        (Feed filePath feedPath feedMeta feedContent) ->
          buildFeed struct filePath feedPath feedMeta feedContent
  where
    buildDir =
      path </> "build"

    buildPage struct templates filePath pageMeta pageContent = do
      pageAsTemplate <- compileMustacheFile' filePath pageContent
      let pageMetaWithStructure = M.insert "_structure" struct pageMeta
          renderedPage = LT.toStrict $ renderMustache pageAsTemplate (Aeson.Object pageMetaWithStructure)
      pageHTML <- commonMarkToHTML filePath renderedPage
      let pageMetaWithHTML = M.insert "_html" (Aeson.String pageHTML) pageMetaWithStructure
          pageOutput = case getPageTemplate pageMeta templates of
            Just t -> LT.toStrict $ renderMustache t (Aeson.Object pageMetaWithHTML)
            Nothing -> pageHTML
          pageDir = case M.lookup "_path" pageMeta of
            Just (Aeson.String p) -> buildDir <> fromText p
            _ -> bug "page _path is not set or not a string"
      createDirectoryIfMissing True pageDir
      BS.writeFile (pageDir </> "index.html") (encodeUtf8 pageOutput)

    buildFeed struct filePath feedPath feedMeta feedContent = do
      feedAsTemplate <- compileMustacheFile' filePath feedContent
      let feedMetaWithStructure = M.insert "_structure" struct feedMeta
          feedOutput = LT.toStrict $ renderMustache feedAsTemplate (Aeson.Object feedMetaWithStructure)
          feedOutputPath = buildDir <> fromText feedPath
      createDirectoryIfMissing True (dropFileName feedOutputPath)
      BS.writeFile feedOutputPath (encodeUtf8 feedOutput)

    copyRecursively from to =
      walkCopy from
      where
        walkCopy f =
          doesDirectoryExist f >>= \case
            False -> copyFile f t
            True -> do
              createDirectoryIfMissing False t
              listDirectory f >>= traverse_ (walkCopy . (f </>))
          where
            t = to </> makeRelative from f

    compileMustacheFile' filePath content =
      case compileMustacheText "" (LT.fromStrict content) of
        Right t -> pure t
        Left err -> error $ "can't process file: path='" <> toText filePath <> "' err='" <> toText (show err) <> "'"

    commonMarkToHTML pagePath pageContent =
      case commonmark pagePath pageContent of
        Right (html :: Html ()) -> pure $ LT.toStrict $ renderHtml html
        Left err -> error $ "can't convert to HTML: path='" <> toText pagePath <> "' err='" <> toText (show err) <> "'"

    getPageTemplate meta templates =
      let templateName =
            case M.lookup "_template" meta of
              Just (Aeson.String tname) -> PName tname
              _ -> PName "default"
       in case Map.lookup templateName (templateCache templates) of
            Just _ -> Just templates {templateActual = templateName}
            Nothing -> Nothing

readContent root = do
  buildTime <- toText . iso8601Show <$> getCurrentTime
  doesDirectoryExist contentDir >>= \case
    False ->
      error $ "directory does not exist: " <> toText contentDir
    True ->
      walk buildTime contentDir M.empty
        <&> first (fromMaybe (bug "site structure is not built") . M.lookup "_items")
  where
    walk buildTime path meta
      | path `L.elem` ignoredPaths =
        pure (M.empty, [])
      | otherwise =
        doesDirectoryExist path >>= \case
          True ->
            listDirectory path >>= processCollection
          False -> do
            fileExists <- doesFileExist path
            if
                | fileExists && ".md" `L.isSuffixOf` path -> processPage
                | fileExists && "feed.xml" == takeFileName path -> processFeed
                | otherwise -> pure (M.empty, [])
      where
        processCollection children = do
          collectionMeta <- readMetadata (path </> "metadata.json")
          let mergedMeta = M.union collectionMeta meta
              comparator = getCollectionComparator mergedMeta
          (structs, pages) <-
            traverse (\c -> walk buildTime (path </> c) mergedMeta) (L.sort children)
              <&> (first (L.sortBy comparator . removeDuplicates . L.filter (not . M.null)) . L.unzip)
          let predefMeta =
                M.fromList
                  [ ("_name", Aeson.String (toText $ takeBaseName path)),
                    ("_is_page", Aeson.Bool False),
                    ("_is_collection", Aeson.Bool True),
                    ("_items", Aeson.Array (V.fromList $ fmap Aeson.Object structs)),
                    ("_build_time", Aeson.String buildTime)
                  ]
          pure (M.union predefMeta collectionMeta, L.concat pages)

        processPage = do
          (pageMeta, pageContent) <- readPage path
          let pagePath =
                if "index.md" `L.isSuffixOf` path
                  then dropFileName $ stripContentDir path
                  else dropExtension (stripContentDir path) <> "/"
              mergedMeta = M.union pageMeta meta
              predefMeta =
                M.fromList
                  [ ("_name", Aeson.String (toText $ takeBaseName path)),
                    ("_path", Aeson.String (toText pagePath)),
                    ("_is_page", Aeson.Bool True),
                    ("_is_collection", Aeson.Bool False),
                    ("_build_time", Aeson.String buildTime)
                  ]
          pure (M.union predefMeta pageMeta, [Page path (M.union predefMeta mergedMeta) (decodeUtf8 pageContent)])

        processFeed = do
          feedContent <- decodeUtf8 <$> BS.readFile path
          let feedPath = toText $ stripContentDir path
              predefMeta = M.fromList [("_build_time", Aeson.String buildTime)]
          pure (M.empty, [Feed path feedPath (M.union predefMeta meta) feedContent])

        getCollectionComparator m =
          let (fname, rev) =
                case (M.lookup "_sort_by" m, M.lookup "_reverse" m) of
                  (Just (Aeson.String fname'), Just (Aeson.Bool rev')) -> (fname', rev')
                  (Just (Aeson.String fname'), _) -> (fname', False)
                  (_, Just (Aeson.Bool rev')) -> ("_name", rev')
                  _ -> ("_name", False)
              key = Key.fromText fname
           in \a b ->
                let a' = M.lookup key a
                    b' = M.lookup key b
                 in if rev then compare (Down a') (Down b') else compare a' b'

        removeDuplicates structs =
          remove structs
          where
            remove [] =
              []
            remove (x : xs)
              | M.lookup "_is_collection" x == Just (Aeson.Bool True) =
                x : remove xs
              | L.length (L.findIndices (\a -> M.lookup "_name" a == M.lookup "_name" x) structs) > 1 =
                remove xs
              | otherwise =
                x : remove xs

    contentDir =
      root </> "content"

    stripContentDir =
      fromMaybe (bug "stripContentDir used on a file outside content directory") . L.stripPrefix contentDir

    ignoredPaths =
      (contentDir </>) <$> ["static", "static.md"]

    readMetadata path =
      doesFileExist path >>= \case
        False ->
          pure M.empty
        True ->
          Aeson.eitherDecodeFileStrict' path >>= \case
            Right meta -> pure meta
            Left err -> error $ "can't parse metadata: path='" <> toText path <> "' err='" <> toText err <> "'"

    readPage path = do
      BS.readFile path <&> splitPage >>= \case
        (Just frontMatter, pageData) ->
          case Aeson.fromJSON <$> Aeson.eitherDecodeStrict' frontMatter of
            Right (Aeson.Success meta) ->
              pure (meta, pageData)
            Right (Aeson.Error err) ->
              error $ "invalid page metadata: path='" <> toText path <> "' err='" <> toText err <> "'"
            Left err ->
              error $ "unable to parse metadata: path='" <> toText path <> "' err='" <> toText err <> "'"
        (Nothing, pageData) ->
          pure (M.empty, pageData)
      where
        splitPage pageData =
          let (beforeFrontMatter, frontMatterAndRest) = BS.breakSubstring "---" pageData
              (frontMatter, rest) = BS.breakSubstring "..." frontMatterAndRest
           in if BS.null beforeFrontMatter
                then (Just (BS.drop 3 frontMatter), BS.drop 3 rest)
                else (Nothing, pageData)

data Content
  = -- | Page filePath metadata content
    Page FilePath Aeson.Object Text
  | -- | Feed filePath feedPath metadata content
    Feed FilePath Text Aeson.Object Text
  deriving (Eq, Show)

parseTemplates root = do
  templates <-
    listDirectory templatesDir
      <&> L.filter (".html" `L.isSuffixOf`)
      <&> fmap (templatesDir </>)
      >>= filterM doesFileExist
  foldr' (<>) emptyTemplate <$> traverse compileMustacheFile templates
  where
    templatesDir = root </> "templates"
    emptyTemplate = fromRight (bug "empty template is not parsed") $ compileMustacheText "" ""